import sys

from qtpy import QtCore, QtGui, QtWidgets

import qtawesome


# TODO: Set icon colour and copy code with color kwarg

DEFAULT_VIEW_COLUMNS = 5
VIEW_COLUMNS_OPTIONS = [5, 8, 10, 15, 20, 25, 30]
AUTO_SEARCH_TIMEOUT = 500
ALL_COLLECTIONS = "All"


class IconBrowser(QtWidgets.QMainWindow):
    """
    A small browser window that allows the user to search through all icons from
    the available version of QtAwesome.  You can also copy the name and python
    code for the currently selected icon.
    """

    def __init__(self):
        super().__init__()

        qtawesome._instance()
        fontMaps = qtawesome._resource["iconic"].charmap

        iconNames = []
        for fontCollection, fontData in fontMaps.items():
            for iconName in fontData:
                iconNames.append("%s.%s" % (fontCollection, iconName))

        self.setMinimumSize(300, 300)
        self.setWindowTitle("QtAwesome Icon Browser")
        self.setWindowIcon(qtawesome.icon("fa5s.icons"))

        self._filterTimer = QtCore.QTimer(self)
        self._filterTimer.setSingleShot(True)
        self._filterTimer.setInterval(AUTO_SEARCH_TIMEOUT)
        self._filterTimer.timeout.connect(self._updateFilter)

        model = IconModel()
        model.setStringList(sorted(iconNames))

        self._proxyModel = QtCore.QSortFilterProxyModel()
        self._proxyModel.setSourceModel(model)
        self._proxyModel.setFilterCaseSensitivity(QtCore.Qt.CaseInsensitive)

        self._listView = IconListView(DEFAULT_VIEW_COLUMNS, parent=self)
        self._listView.setUniformItemSizes(True)
        self._listView.setViewMode(QtWidgets.QListView.IconMode)
        self._listView.setModel(self._proxyModel)
        self._listView.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self._listView.doubleClicked.connect(self._copyIconText)
        self._listView.selectionModel().selectionChanged.connect(self._updateNameField)

        toolbar = QtWidgets.QHBoxLayout()

        # Filter section
        self._comboFont = QtWidgets.QComboBox(self)
        self._comboFont.setToolTip(
            "Select the font prefix whose icons will be included in the filtering."
        )
        self._comboFont.setMaximumWidth(75)
        self._comboFont.addItems([ALL_COLLECTIONS] + sorted(fontMaps.keys()))
        self._comboFont.currentIndexChanged.connect(self._triggerImmediateUpdate)
        toolbar.addWidget(self._comboFont)

        self._lineEditFilter = QtWidgets.QLineEdit(self)
        self._lineEditFilter.setToolTip("Filter icons by name")
        self._lineEditFilter.setMaximumWidth(200)
        self._lineEditFilter.setToolTip("Filter icons by name")
        self._lineEditFilter.setAlignment(QtCore.Qt.AlignLeft)
        self._lineEditFilter.textChanged.connect(self._triggerDelayedUpdate)
        self._lineEditFilter.returnPressed.connect(self._triggerImmediateUpdate)
        self._lineEditFilter.setClearButtonEnabled(True)
        toolbar.addWidget(self._lineEditFilter, stretch=10)

        # Icon name section
        self._nameField = QtWidgets.QLineEdit(self)
        self._nameField.setPlaceholderText(
            "Full identifier of the currently selected icon"
        )
        self._nameField.setAlignment(QtCore.Qt.AlignCenter)
        self._nameField.setReadOnly(True)
        self._nameField.setMaximumWidth(250)
        fnt = self._nameField.font()
        fnt.setFamily("monospace")
        fnt.setBold(True)
        self._nameField.setFont(fnt)
        toolbar.addWidget(self._nameField, stretch=10)

        self._copyButton = QtWidgets.QPushButton("Copy Name", self)
        self._copyButton.setToolTip(
            "Copy selected icon full identifier to the clipboard"
        )
        self._copyButton.clicked.connect(self._copyIconText)
        self._copyButton.setDisabled(True)
        toolbar.addWidget(self._copyButton)
        toolbar.addStretch(1)

        # Style section
        self._comboStyle = QtWidgets.QComboBox(self)
        self._comboStyle.setToolTip(
            "Select color palette for the icons and the icon browser"
        )
        self._comboStyle.addItem(qtawesome.styles.DEFAULT_DARK_PALETTE, 0)
        self._comboStyle.addItem(qtawesome.styles.DEFAULT_LIGHT_PALETTE, 1)
        self._comboStyle.currentTextChanged.connect(self._updateStyle)
        toolbar.addWidget(self._comboStyle)

        # Display (columns number) section
        self._comboColumns = QtWidgets.QComboBox(self)
        self._comboColumns.setToolTip(
            "Select number of columns the icons list is showing"
        )
        for num_columns in VIEW_COLUMNS_OPTIONS:
            self._comboColumns.addItem(str(num_columns), num_columns)
        self._comboColumns.setCurrentIndex(
            self._comboColumns.findData(DEFAULT_VIEW_COLUMNS)
        )
        self._comboColumns.currentTextChanged.connect(self._updateColumns)
        toolbar.addWidget(self._comboColumns)

        # Layout
        lyt = QtWidgets.QVBoxLayout()
        lyt.addLayout(toolbar)
        lyt.addWidget(self._listView)

        frame = QtWidgets.QFrame(self)
        frame.setLayout(lyt)

        self.setCentralWidget(frame)

        self.setTabOrder(self._comboFont, self._lineEditFilter)
        self.setTabOrder(self._lineEditFilter, self._comboStyle)
        self.setTabOrder(self._comboStyle, self._listView)
        self.setTabOrder(self._listView, self._nameField)
        self.setTabOrder(self._nameField, self._copyButton)
        self.setTabOrder(self._copyButton, self._comboFont)

        # Shortcuts
        QtWidgets.QShortcut(
            QtGui.QKeySequence(QtCore.Qt.Key_Return),
            self,
            self._copyIconText,
        )
        QtWidgets.QShortcut(
            QtGui.QKeySequence("Ctrl+F"),
            self,
            self._lineEditFilter.setFocus,
        )

        self._lineEditFilter.setFocus()

        geo = self.geometry()

        # QApplication.desktop() has been removed in Qt 6.
        # Instead, QGuiApplication.screenAt(QPoint) is supported
        # in Qt 5.10 or later.
        try:
            screen = QtGui.QGuiApplication.screenAt(QtGui.QCursor.pos())
            centerPoint = screen.geometry().center()
        except AttributeError:
            desktop = QtWidgets.QApplication.desktop()
            screen = desktop.screenNumber(desktop.cursor().pos())
            centerPoint = desktop.screenGeometry(screen).center()

        geo.moveCenter(centerPoint)
        self.setGeometry(geo)
        self._updateStyle(self._comboStyle.currentText())

    def _updateStyle(self, text: str):
        _app = QtWidgets.QApplication.instance()
        if text == qtawesome.styles.DEFAULT_DARK_PALETTE:
            qtawesome.reset_cache()
            qtawesome.dark(_app)
        else:
            qtawesome.reset_cache()
            qtawesome.light(_app)

    def _updateColumns(self):
        self._listView.setColumns(self._comboColumns.currentData())

    def _updateFilter(self):
        """
        Update the string used for filtering in the proxy model with the
        current text from the line edit.
        """
        reString = ""

        group = self._comboFont.currentText()
        if group != ALL_COLLECTIONS:
            reString += r"^%s\." % group

        searchTerm = self._lineEditFilter.text()
        if searchTerm:
            reString += ".*%s.*$" % searchTerm

        # QSortFilterProxyModel.setFilterRegExp has been removed in Qt 6.
        # Instead, QSortFilterProxyModel.setFilterRegularExpression is
        # supported in Qt 5.12 or later.
        try:
            self._proxyModel.setFilterRegularExpression(reString)
        except AttributeError:
            self._proxyModel.setFilterRegExp(reString)

    def _triggerDelayedUpdate(self):
        """
        Reset the timer used for committing the search term to the proxy model.
        """
        self._filterTimer.stop()
        self._filterTimer.start()

    def _triggerImmediateUpdate(self):
        """
        Stop the timer used for committing the search term and update the
        proxy model immediately.
        """
        self._filterTimer.stop()
        self._updateFilter()

    def _copyIconText(self):
        """
        Copy the name of the currently selected icon to the clipboard.
        """
        indexes = self._listView.selectedIndexes()
        if not indexes:
            return

        clipboard = QtWidgets.QApplication.instance().clipboard()
        clipboard.setText(indexes[0].data())

    def _updateNameField(self):
        """
        Update field to the name of the currently selected icon.
        """
        indexes = self._listView.selectedIndexes()
        if not indexes:
            self._nameField.setText("")
            self._copyButton.setDisabled(True)
            return

        self._nameField.setText(indexes[0].data())
        self._copyButton.setDisabled(False)


class IconListView(QtWidgets.QListView):
    """
    A QListView that scales it's grid size to ensure the same number of
    columns are always drawn.
    """

    def __init__(self, columns, parent=None):
        super().__init__(parent)
        self._columns = columns
        self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOn)

    def setColumns(self, cols):
        """
        Set columns number and resize.
        """
        self._columns = cols
        self._resize()

    def _resize(self):
        """
        Set grid and icon size taking into account the number of columns.
        """

        width = self.viewport().width() - 30
        # The minus 30 above ensures we don't end up with an item width that
        # can't be drawn the expected number of times across the view without
        # being wrapped. Without this, the view can flicker during resize
        tileWidth = width / self._columns
        iconWidth = int(tileWidth * 0.8)
        # tileWidth needs to be an integer for setGridSize
        tileWidth = int(tileWidth)

        self.setGridSize(QtCore.QSize(tileWidth, tileWidth))
        self.setIconSize(QtCore.QSize(iconWidth, iconWidth))

    def resizeEvent(self, event):
        """
        Re-implemented to resize view following number of columns available.
        """
        self._resize()
        return super().resizeEvent(event)


class IconModel(QtCore.QStringListModel):
    def __init__(self):
        super().__init__()

    def flags(self, index):
        return QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable

    def data(self, index, role):
        """
        Re-implemented to return the icon for the current index.

        Parameters
        ----------
        index : QtCore.QModelIndex
        role : int

        Returns
        -------
        Any
        """
        if role == QtCore.Qt.DecorationRole:
            iconString = self.data(index, role=QtCore.Qt.DisplayRole)
            return qtawesome.icon(iconString)
        return super().data(index, role)


def run():
    """
    Start the IconBrowser and block until the process exits.
    """
    app = QtWidgets.QApplication([])
    qtawesome.dark(app)

    browser = IconBrowser()
    browser.show()

    sys.exit(app.exec_())


if __name__ == "__main__":
    run()



r"""
Tests for QtAwesome.
"""

# Standard library imports
import collections
import os
import subprocess
import sys

# Test Library imports
import pytest

# Local imports
from qtawesome.iconic_font import IconicFont
import qtawesome as qta


def test_segfault_import():
    output_number = subprocess.call(
        sys.executable + ' -c "import qtawesome ; qtawesome.icon()"', shell=True
    )
    assert output_number == 0


def test_unique_font_family_name(qtbot):
    """
    Test that each font used by qtawesome has a unique name. If this test
    fails, this probably means that you need to rename the family name of
    some fonts. Please see PR #98 for more details on why it is necessary and
    on how to do this.

    Regression test for Issue #107
    """
    resource = qta._instance()
    assert isinstance(resource, IconicFont)

    # Check that the fonts were loaded successfully.
    fontnames = resource.fontname.values()
    assert fontnames

    # Check that qtawesome does not load fonts with duplicate family names.
    duplicates = [
        fontname
        for fontname, count in collections.Counter(fontnames).items()
        if count > 1
    ]
    assert not duplicates


@pytest.mark.skipif(os.name != "nt", reason="Only meant for Windows")
def test_bundled_font_user_installation():
    """
    Test that the bundled fonts are being installed on Windows for current user.

    See spyder-ide/qtawesome#167 and spyder-ide/spyder#18642
    """
    qta._instance()
    fonts_expected = [
        font_filename
        for _prefix, font_filename, _charmap_filename in qta._BUNDLED_FONTS
    ]
    fonts_command = [
        "powershell.exe",
        r'Get-ItemProperty "HKCU:\SOFTWARE\Microsoft\Windows NT\CurrentVersion\Fonts"',
    ]
    fonts_result = (
        subprocess.run(fonts_command, capture_output=True, check=True, text=True)
        .stdout.replace("\n", "")
        .replace(" ", "")
    )
    for font_filename in fonts_expected:
        assert font_filename in fonts_result


def test_get_fonts_info():
    """
    Test that you can get the info of all the bundled fonts.
    """
    fonts_expected = [
        font_filename
        for _prefix, font_filename, _charmap_filename in qta._BUNDLED_FONTS
    ]
    fonts_root_dir, fonts_list = qta.get_fonts_info()
    assert os.path.normcase(fonts_root_dir) == os.path.normcase(
        os.path.join(
            os.path.dirname(os.path.dirname(os.path.realpath(__file__))), "fonts"
        )
    )
    assert set(fonts_list) == set(fonts_expected)


@pytest.mark.skipif(os.name != "nt", reason="Only meant for Windows")
def test_bundled_font_system_installation():
    """
    Test that the bundled fonts can be installed on Windows for all users.

    See spyder-ide/qtawesome#244

    Notes
    -----
    * When running this test, it's possible that a prompt for privileges
      may appear (UAC prompt).
    """
    qta.install_bundled_fonts_system_wide()
    fonts_expected = [
        font_filename
        for _prefix, font_filename, _charmap_filename in qta._BUNDLED_FONTS
    ]
    assert len(fonts_expected) == len(qta._BUNDLED_FONTS)
    fonts_command = [
        "powershell.exe",
        r'Get-ItemProperty "HKLM:\SOFTWARE\Microsoft\Windows NT\CurrentVersion\Fonts"',
    ]
    fonts_result = (
        subprocess.run(fonts_command, capture_output=True, check=True, text=True)
        .stdout.replace("\n", "")
        .replace(" ", "")
    )
    for font_filename in fonts_expected:
        assert font_filename in fonts_result


def test_font_load_from_system_fonts(monkeypatch):
    """
    Test that the bundled fonts can be accessed from the system fonts folder on
    Windows.

    Notes
    -----
    * This test ensures that the logic to load fonts from the system fonts
      only affects Windows even when it is being forced.
    * When running this test, it's possible that a prompt for privileges may
      appear (UAC prompt).
    """
    qta.install_bundled_fonts_system_wide()
    with monkeypatch.context() as m:
        qta._resource["iconic"] = None
        m.setenv("QTA_FORCE_SYSTEM_FONTS_LOAD", "true")
        qta._instance()


def test_the_app_exits_gracefully_after_drawing_a_path():
    """
    Test that the app finishes gracefully when the `draw` parameter is `path`.

    See spyder-ide/qtawesome#280

    Notes
    -----
    * This `draw="path"` mode causes the icon to be drawn via `QRawFont`.
      The font is cached for every `QThread`.
      The cache is cleared when the current thread is over.
      If the thread is the main one, this may cause a segfault when the app ends.
    * The issue does not crash the app immediately,
      but causes a segfault at the end only.
      So, we start a new app for it to end within the test.
    """
    ret_code = subprocess.call(
        [
            sys.executable,
            "-c",
            ";".join(
                (
                    "from qtpy.QtWidgets import QApplication",
                    "from qtawesome import IconWidget",
                    "app = QApplication([])",
                    "IconWidget('fa5s.egg', draw='path')",
                )
            ),
        ],
    )
    assert ret_code == 0


if __name__ == "__main__":
    pytest.main()
